/* cfengine for GNU
 
        Copyright (C) 1995
        Free Software Foundation, Inc.
 
   This file is part of GNU cfengine - written and maintained 
   by Mark Burgess, Dept of Computing and Engineering, Oslo College,
   Dept. of Theoretical physics, University of Oslo
 
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

*/


/*******************************************************************/
/*                                                                 */
/* Process handling                                                */
/*                                                                 */
/*******************************************************************/


#include "cf.defs.h"
#include "cf.extern.h"

int DEADLOCK = false;
FILE *PIPE;

/*******************************************************************/

LoadProcessTable(procdata,psopts)

struct Item **procdata;
char *psopts;

{ FILE *pp;
  char pscomm[maxlinksize];

sprintf(pscomm,"%s %s",VPSCOMM[VSYSTEMHARDCLASS],psopts);
  
if ((pp = cfpopen(pscomm,"r")) == NULL)
   {
   sprintf(OUTPUT,"Couldn't open the process list with command %s\n",pscomm);
   CfLog(cferror,OUTPUT,"popen");
   return false;
   }

while (!feof(pp))
   {
   ReadLine(VBUFF,bufsize,pp);
   AppendItem(procdata,VBUFF,"");
   }

cfpclose(pp);
return true;
}

/*******************************************************************/

DoProcessCheck(pp,procdata)

struct Process *pp;
struct Item *procdata;

{ struct Item *ip;
  char fastmap[(1 << BYTEWIDTH)];
  struct re_pattern_buffer buf;
  struct re_registers regs;
  char *err, line[bufsize], *sp;
  int match, pid=-1, ret, matches=0;
  int BreakPipe();
  FILE *pipe;

Debug2("Looking for process %s\n",pp->expr);

buf.allocated = 0;
buf.buffer = NULL;
buf.fastmap = fastmap;
buf.translate = 0;

if ((err = re_compile_pattern(pp->expr,strlen(pp->expr),&buf)) != 0)
   {
   sprintf(OUTPUT,"Regular expression error for process pattern %s\n",pp->expr);
   CfLog(cferror,OUTPUT,"");
   sprintf(OUTPUT,"%s\n",err);
   CfLog(cferror,OUTPUT,"");
   return;
   }

re_compile_fastmap (&buf);

for (ip = procdata; ip != NULL; ip=ip->next)
   {
   if ((match = re_search(&buf,ip->name,strlen(ip->name),0,strlen(ip->name),&regs)) >= 0)
      {
      pid = -1;

      for (sp = ip->name; *sp != '\0'; sp++) /* if first field contains alpha, skip */
	 {
	 while (!isdigit(*sp) && (*sp != '\0'))
	    {
	    sp++;
	    }
      
	 sscanf(sp,"%d",&pid);

	 if (pid != -1)
	    {
	    break;
	    }
	 }  
      
      if (pid == -1)
	 {
	 sprintf(OUTPUT,"Unable to extract pid while looking for %s\n",pp->expr);
	 CfLog(cfverbose,OUTPUT,"");
	 continue;
	 }

      Debug2("Found matching pid %d\n",pid);

      matches++;
      
      if (pid == 1 && pp->signal == cfhup)
	 {
	 Verbose("(Okay to send HUP to init)\n");
	 }
      else if (pid < 4)
	 {
	 printf("%s: will not signal or restart processes 0,1,2,3\n",VPREFIX);
	 printf("%s: occurred while looking for %s\n",VPREFIX,pp->expr);
	 }

      if (pp->action == 'w')
	 {
	 sprintf(OUTPUT,"Process warning, found: %s\n",ip->name);
	 CfLog(cfinform,OUTPUT,"");
	 continue;
	 }

      if (pp->signal != cfnosignal)
	 {
	 sprintf(OUTPUT,"Signalling process %d (%s) with %s\n",pid,pp->expr,SIGNALS[pp->signal]);
	 CfLog(cfinform,OUTPUT,"");

	 if (!DONTDO)
	    {
	    if ((ret = kill((pid_t)pid,pp->signal)) < 0)
	       {
	       sprintf(OUTPUT,"Couldn't send signal to pid %d\n",pid);
	       CfLog(cferror,OUTPUT,"kill");

	       continue;
	       }

	    if ((pp->signal == cfkill || pp->signal == cfterm) && ret >= 0)
	       {
	       sprintf(OUTPUT,"Killed: %s\n",ip->name);
	       CfLog(cfinform,OUTPUT,"");
	       }
	    }
	 }

      }
   else if (match < -1)
      {
      sprintf(OUTPUT,"Regular expression internal error [%s]\n",pp->expr);
      CfLog(cferror,OUTPUT,"");
      }
   }

if (pp->matches >= 0)
   {
   AddMultipleClasses(pp->defines);

   switch (pp->comp)
      {
      case '=': if (matches != (int)pp->matches)
                   {
	    	   sprintf(OUTPUT,"%d processes matched %s (should be %d)\n",matches,pp->expr,pp->matches);
		   CfLog(cfinform,OUTPUT,"");
		   }
                break;
		
      case '>': if (matches <= (int)pp->matches)
	           {
	    	   sprintf(OUTPUT,"%d processes matched %s (should be >%d)\n",matches,pp->expr,pp->matches);
		   CfLog(cfinform,OUTPUT,"");
	           } 
                break;
		
      case '<': if (matches >= (int)pp->matches)
	           {
	    	   sprintf(OUTPUT,"%d processes matched %s (should be <%d)\n",matches,pp->expr,pp->matches);
		   CfLog(cfinform,OUTPUT,"");
	           } 
      }
   }

if (strlen(pp->restart) != 0)
   {
   if ((matches != 0) && (pp->signal != cfkill) && (pp->signal != cfterm))
      {
      Verbose("Matches found for %s - no restart sequence\n",pp->expr);
      return;
      }

   sprintf(OUTPUT,"Executing shell command: %s\n",pp->restart);
   CfLog(cfinform,OUTPUT,"");

   if (DONTDO)
      {
      return;
      }

   if (pp->useshell == 'y')
      {
      if ((PIPE = popen(pp->restart,"r")) == NULL)
	 {
	 sprintf(OUTPUT,"Process restart execution failed on %s\n",pp->restart);
	 CfLog(cferror,OUTPUT,"popen");
	 return;
	 }
      }
   else
      {
      if ((PIPE = cfpopen(pp->restart,"r")) == NULL)
	 {
	 sprintf(OUTPUT,"Process restart execution failed on %s\n",pp->restart);
	 CfLog(cferror,OUTPUT,"popen");
	 return;
	 }
      }

   DEADLOCK = false;   
   signal(SIGALRM,(void *)BreakPipe);
   alarm(30);
   
   while (!feof(PIPE))
      {
      ReadLine(line,bufsize,PIPE);

      if (feof(PIPE) || ferror(PIPE))
	 {
	 break;
	 }
      
      if (strstr(line,"cfengine-die"))
	 {
	 break;
	 }

      /* patch for ERESTARTSYSTEM bug in popen */

      if (strstr(line,"cfd: start") || DEADLOCK)
	 {
         printf("");
	 break;
	 }
      
      sprintf("Restart: %s",line);
      CfLog(cfinform,OUTPUT,"");
      }

   alarm(0);
   signal(SIGALRM,SIG_DFL);

   if (pp->useshell == 'y')
      {
      pclose(PIPE);
      }
   else
      {
      cfpclose(PIPE);
      }
   
   sprintf(OUTPUT,"(Done with %s)\n",pp->restart);
   CfLog(cfinform,OUTPUT,"");
   }
}

/*******************************************************************/

BreakPipe()

{
Verbose("%s: Workaroud on ERESTARTSYS popen() zombie deadlock bug\n",VPREFIX);
ungetc('\n',PIPE);
ungetc(EOF,PIPE);
DEADLOCK = true;
}
